import datetime
from distutils.util import strtobool
import time
from uuid import uuid4
from loguru import logger
from time import sleep
from typing import Union

import pandas as pd
import sys
import json
import os
import requests
import getpass
import warnings
from requests_oauthlib import OAuth2Session
from oauthlib.oauth2 import OAuth2Error, BackendApplicationClient

from .fileimport import FileImport
from .automation import Automation
from .schedule import Schedule
from .programming import Programming
from .authorization import Authorization
from .email import Email

from .utils.ut import Utils, Structure, Defaults
from .utils.ut_log import LogUtils
from .utils.DeviceClient import DeviceClient
from .utils.ut_auth import AuthData
from .timeseries import TimeSeries


class TechStack():
    """
    Initializes a Seven2one TechStack client.

    Parameters:
    ----------
    host: str
        The host domain or ip address, e.g. 'app.organisation.com'
    client_id: str = None
        The client id belonging to the apps and specific for the system. If None, it needs to
        be entered interactively.
    service_account_name: str
        A techstack service account name. Leave blank to login interactively with user.
    service_account_secret: str = None
        The service account secret belonging to the service account and specific for the system. If None, it needs to
        be entered interactively.
    proxies: dict = None
        Provide one or more proxy addresses, e.g.: {'https':'159.98.7.123'} 
    usePorts: bool = False
        A developer feature: if True, ports with prefix '8' for a
        developer ComposeStack will be used.
    timeZone: str = None
        A time zone provided in IANA or isoformat (e.g. 'Europe/Berlin' or 'CET').
        Defaults to the local timezone. 
    dateTimeOffset: bool = True
        Choose, if a Timestamp should be displayed as time zone naive (dateTimeOffset = False) 
        or with Offset information. 
    raiseException: bool = False
        In default mode (False) exceptions will be avoided to provide better user experience.
        True is recommended in automated processes to avoid silent failures.
    copyGraphQLString: bool = False
        A developer feature: if True, with each method execution, 
        the graphQL string is copied to the clipboard (Windows only).

    Examples:
    >>> client = TechStack('app.orga.com/', client_id='...')
    >>> client = TechStack('app.orga.com/', client_id='...', service_account_name='my-service', service_account_secret='...')) 
    """

    def __init__(
        self,
        host: str,
        client_id: str = None,
        service_account_name: str = None,
        service_account_secret: str = None,
        proxies: dict = None,
        usePorts: bool = False,
        timeZone: str = None,
        dateTimeOffset: bool = True,
        raiseException: bool = False,
        copyGraphQLString: bool = False
    ) -> object:

        self.raiseException = raiseException
        warnings.filterwarnings(
            'ignore', message="The zone attribute is specific to pytz")

        loglevelLocal = os.getenv("LOGLEVEL", "WARNING")
        loglevelServer = os.getenv("LOGLEVEL_SERVER", "ERROR")
        logToServer = strtobool(os.getenv("LOG_TO_SERVER", 'True').lower())
        sessionId = str(uuid4())

        try:
            logger.remove()

            if loglevelLocal in ['DEBUG', 'TRACE']:
                logger_format = "{level:<10} {time} {module}.{function} {line}: {message}"
                logger_diagnose = True
                logger_backtrace = True
            else:
                logger_format = "{level:<10} {time} {message}"
                logger_diagnose = False
                logger_backtrace = False

            def stdout_log_filter(record):
                return record["level"].no < logger.level('ERROR').no

            logger.add(sys.stdout, format=logger_format, level=loglevelLocal, catch=True,
                       diagnose=logger_diagnose, backtrace=logger_backtrace, filter=stdout_log_filter)
            logger.add(sys.stderr, format=logger_format, level='ERROR',
                       catch=True, diagnose=logger_diagnose, backtrace=logger_backtrace)
        except:
            pass

        if os.name == 'nt':
            logger.debug('Detected Windows, enabling pyperclip')
        else:
            logger.debug(f"Detected platform: {os.name}")

        if usePorts == False:
            idP_url = f'https://authentik.{host}'
            dynEndpoint = f'https://{host}/dynamic-objects/graphql/'
            automationEndpoint = f'https://{host}/automation/graphql/'
            scheduleEndpoint = f'https://{host}/schedule/graphql/'
            programmingEndpoint = f'https://{host}/programming/graphql/'
            tsGatewayEndpoint = f'https://{host}/timeseries/graphql/'
            logEndpoint = f'https://{host}/logging/loki/api/v1/push'
            authzEndpoint = f'https://{host}/authz/graphql/'
            emailEndpoint = f'https://{host}/emailservice/Email'
        else:
            idP_url = f'http://{host}:8044'
            dynEndpoint = f'http://{host}:8050/graphql/'
            automationEndpoint = f'http://{host}:8120/graphql/'
            scheduleEndpoint = f'http://{host}:8130/graphql/'
            programmingEndpoint = f'http://{host}:8140/graphql/'
            tsGatewayEndpoint = f'http://{host}:8195/graphql/'
            logEndpoint = f'http://{host}:8175/loki/api/v1/push'
            authzEndpoint = f'http://{host}:8030/graphql/'
            emailEndpoint = f'http://{host}:8240/Email'

        if os.getenv("IDENDITYPROVIDER_URL") != None:
            idP_url = os.getenv("IDENDITYPROVIDER_URL")
        if os.getenv("DYNAMIC_OBJECTS_ENDPOINT") != None:
            dynEndpoint = os.getenv("DYNAMIC_OBJECTS_ENDPOINT")
        if os.getenv("AUTOMATION_ENDPOINT") != None:
            automationEndpoint = os.getenv("AUTOMATION_ENDPOINT")
        if os.getenv("SCHEDULE_ENDPOINT") != None:
            scheduleEndpoint = os.getenv("SCHEDULE_ENDPOINT")
        if os.getenv("PROGRAMMING_ENDPOINT") != None:
            programmingEndpoint = os.getenv("PROGRAMMING_ENDPOINT")
        if os.getenv("TIMESERIES_ENDPOINT") != None:
            tsGatewayEndpoint = os.getenv("TIMESERIES_ENDPOINT")
        if os.getenv("LOGGING_ENDPOINT") != None:
            logEndpoint = os.getenv("LOGGING_ENDPOINT")
        if os.getenv("AUTHORIZATION_ENDPOINT") != None:
            authzEndpoint = os.getenv("AUTHORIZATION_ENDPOINT")

        if usePorts == True:
            os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1'

        token_endpoint = '/application/o/token/'
        token_url = f'{idP_url}{token_endpoint}'

        if client_id is None:
            client_id = getpass.getpass('Enter client id: ')

        if service_account_name is None and service_account_secret is None:
            device_endpoint = '/application/o/device/'
            response = requests.request("POST", f'{idP_url}{device_endpoint}', data=f'client_id={client_id}', headers={
                                        'Content-Type': 'application/x-www-form-urlencoded'})
            json_response = json.loads(response.text)
            device_code = json_response['device_code']
            print('Please go to %s and authorize access.' %
                json_response['verification_uri_complete'])
            device_client = DeviceClient(client_id=client_id)
            oauth = OAuth2Session(client=device_client)
            while True:
                try:
                    token = oauth.fetch_token(
                        token_url=token_url, code=device_code, include_client_id=True)
                    break
                except OAuth2Error as e:
                    if e.error == 'authorization_pending':
                        time.sleep(json_response['interval'])
                    else:
                        print(f'Authorization failed: {e.error}')
                        raise
        elif service_account_secret is None:
            service_account_secret = getpass.getpass('Enter secret: ')
        else:
            app_client = BackendApplicationClient(client_id=client_id)
            oauth = OAuth2Session(client=app_client)
            token = oauth.fetch_token(token_url=token_url, 
                                      client_id=client_id, 
                                      username=service_account_name, 
                                      password=service_account_secret)

        if token is None:
            raise Exception('Could not get token')

        auth_data = AuthData(client_id, token_url, token, service_account_name, service_account_secret)

        if logToServer:
            if os.getenv("LOG_SERVER") != None:
                logEndpoint = os.getenv("LOG_SERVER")
            LogUtils._init_logging(
                logEndpoint, auth_data, loglevelServer, sessionId)

        def getLocalTimeZone():
            local_now = datetime.datetime.utcnow().astimezone()
            return local_now.tzinfo.tzname(local_now)
        logger.info(json.dumps({
            "SessionId": sessionId,
            "Configuration": {
                "host": host,
                "usePorts": usePorts,
                "timeZone": timeZone,
                "dateTimeOffset": dateTimeOffset,
                "raiseException": raiseException,
                "copyGraphQLString": copyGraphQLString,
                "logLevel": loglevelLocal,
                "logToServer": logToServer,
                "logLevelServer": loglevelServer
            },
            "System": {
                "timeZone": getLocalTimeZone(),
                "pythonVersion": sys.version,
                "machineName": os.getenv("HOSTNAME"),
                "osName": os.name
            }
        }))
        self.endpoint = dynEndpoint
        self.proxies = proxies
        self.client = Utils._create_client(self, auth_data)
        self.scheme = self.client.introspection
        self.host = host
        self.auth_data = auth_data

        # Defaults:
        if timeZone is None:
            timeZone = Utils._timeZone('local')

        self.defaults = Defaults(
            useDateTimeOffset=dateTimeOffset,
            copyGraphQLString=copyGraphQLString,
            timeZone=timeZone
        )

        # Get scheme
        if self.scheme == None:
            graphQLString = Structure._introspectionQueryString()
            self.scheme = Utils._executeGraphQL(self, graphQLString)

        # Structures
        self.structure = Utils._executeGraphQL(self, Structure.queryStructure)
        self.structure = Structure._fullStructureDict(self.structure)
        self.objects = Structure._fullStructureNT(self.structure)
        self.inventory = Structure._inventoryNT(self.structure)
        self.inventoryProperty = Structure._inventoryPropertyNT(self.structure)

        # Initialize further gateways
        try:
            self.TimeSeries = TimeSeries(
                tsGatewayEndpoint, self, auth_data)
        except Exception as err:
            logger.warning(f"Time series gateway not available")
            logger.debug(f"Reason: {err}")

        try:
            self.Automation = Automation(
                automationEndpoint, self, auth_data)
        except Exception as err:
            logger.warning(f"Automation service not available")
            logger.debug(f"Reason: {err}")

        try:
            self.Schedule = Schedule(scheduleEndpoint, self, auth_data)
        except Exception as err:
            logger.warning(f"Schedule service not available")
            logger.debug(f"Reason: {err}")

        try:
            self.Programming = Programming(
                programmingEndpoint, self, auth_data)
        except Exception as err:
            logger.warning(f"Programming service not available")
            logger.debug(f"Reason: {err}")

        self.FileImport = FileImport(self)
        self.Authorization = Authorization(
            authzEndpoint, self, auth_data)
        self.Email = Email(emailEndpoint, self, auth_data)

        def getVersions():
            versions = [
                self.getVersion(),
                self.TimeSeries.getVersion(),
                self.Authorization.getVersion(),
                self.Automation.getVersion(),
                self.Schedule.getVersion(),
                self.Programming.getVersion()
            ]
            return versions
        logger.opt(lazy=True).debug("Service versions: {}", getVersions)

        return

    def getVersion(self):
        """
        Returns name and version of the responsible micro service
        """

        key = 'dynamicObjectsServiceInfo'
        graphQLString = f'''query version {{ 
            {key} {{
                name
                informationalVersion
            }}
        }}'''
        result = Utils._executeGraphQL(self, graphQLString)

        return f'{result[key]["name"]}: {result[key]["informationalVersion"]}'

    def updateClient(self) -> None:
        """
        Updates the client scheme and structures, e.g. after adding inventories
        or new inventory properties.
        """

        self.client = Utils._create_client(self, self.auth_data)
        graphQLString = Structure._introspectionQueryString()
        self.scheme = Utils._executeGraphQL(self, graphQLString)
        self.structure = Utils._executeGraphQL(self, Structure.queryStructure)
        self.structure = Structure._fullStructureDict(self.structure)
        self.objects = Structure._fullStructureNT(self.structure)
        self.inventory = Structure._inventoryNT(self.structure)
        self.inventoryProperty = Structure._inventoryPropertyNT(self.structure)
        # reninit relevant gateways
        self.TimeSeries = TimeSeries(
            self.TimeSeries.endpoint, self, self.auth_data)

        return

    def inventories(
        self,
        fields: list = None,
        where: str = None,
        orderBy: str = None,
        asc: bool = True
    ) -> pd.DataFrame:
        """
        Returns a DataFrame of existing inventories.

        Parameters:
        ----------
        fields: list=None
            The list of fields to be queried, e.g.
            ['name', 'inventoryId, variant.name']
        where: str=None
            Use a string to add where criteria like 'name eq "meterData"'.
            The argument is momentarily very limited.
        orderBy: str=None
            Select one field to sort by.
        asc: bool=True
            Determines the sort order of items. Set to False to apply descending order.

        Examples:
        >>> inventories()
        >>> inventories(fields=['name', 'inventoryId'], 
                where='city eq "Hamburg"', 
                orderBy='variant', asc=True)
        """

        if fields == None:
            fields = ['name', 'inventoryId', 'variant.name',
                      'historyEnabled', 'hasValidityPeriods', 'isDomainUserType']
            _fields = Utils._queryFields(fields, recursive=True)
        else:
            if type(fields) != list:
                fields = [fields]
            try:
                _fields = Utils._queryFields(fields, recursive=True)
            except:
                Utils._error(
                    self, "Fields must be provided as list, e.g. ['name', 'inventoryId, variant.name']")
                return

        topLevelWhere, _ = Utils._handleWhere(self, where)

        if orderBy != None:
            if asc != True:
                _orderBy = f'order: {{ {orderBy}: DESC }}'
            else:
                _orderBy = f'order: {{ {orderBy}: ASC }}'
        else:
            _orderBy = ''

        graphQLString = f'''query inventories {{
        inventories 
            (pageSize:1000 {_orderBy} {topLevelWhere})
            {{
            {_fields}
            }}
        }}
        '''
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None:
            return

        df = pd.json_normalize(result['inventories'])
        return df

    def items(
        self,
        inventoryName:str, 
        references:bool=False, 
        fields:list=None,
        where:Union[list,tuple,str]=None, 
        orderBy:Union[dict,list,str]=None, 
        asc:Union[list,str]=True, 
        pageSize:int=5000, 
        arrayPageSize:int=50,
        top:int=100000,
        validityDate:str=None,
        allValidityPeriods:bool=False,
        includeSysProperties:bool=False,
        maxRecursionDepth=2
    ) -> pd.DataFrame:
        """
        Returns items of an inventory in a DataFrame.

        Parameters:
        -----------
        inventoryName : str
            The name of the inventory.
        references: bool = False
            If True, items of referenced inventories will be added to the DataFrame. If
            the fields-parameter is used, this parameter is ignored.
        fields: list | str = None
            A list of all properties to be queried. If None, all properties will be queried.
            For referenced items use a '.' between inventory name and property.
            Optional property 'sys_revision': Returns number of updates of an item.
        where: list | tuple | str = None
            Define arguments critera like  'city eq "Berlin" and use lists for AND-combinations and
            tuples for OR-combinations. For references use the format inventory.property as 
            filter criteria.
        orderBy: dict | list | str = None
            Use a dict in the form of {property:'ASC'|'DESC'} or 
            use a list of properties and the asc-argument for sorting direction
        asc: list | bool = True
            Determines the sort order of items. If set to False, a descending order 
            is applied. Use a list, if more properties are selected in orderBy.
        pageSize: int = 5000
            The page size of items that is used to retrieve a large number of items.
        arrayPageSize: int = 50
            The page size of list items that is used to page list items in an inventory item.
        top: int = None
            Returns a restricted set of items oriented at the latest entries.
        includeSysProperties: bool = False
            If True, all system properties available will be queried. If False, 
            only 'sys_inventoryItemId' will be queried by default. Despite of that, any system 
            property can be passed to the fields argument as well.
        validityDate: str = None
            If the queried inventory has validity periods enabled, only items will be returned, 
            which have the given timestamp between sys_validTo and sys_validFrom. Items without
            validity periods are shown as well.
        allValidityPeriods: bool = False
            If True and if the queried inventory has validity periods enabled, all validity 
            periods will be returned. If False, items with validity dates will not be returned.
        maxRecursionDepth: int = 2
            Maximum recursion level following the references to other inventories.

        Example:
        --------
        >>> items('appartments', references=True)
        >>> items(
                'appartments',
                fields=['address', 'owner', 'sys_revision'],
                where=['city == "Berlin"', 'rooms > 2'],
                orderBy={'size':'DESC', price:'ASC'}
                )
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            if inventoryName not in self.inventory:
                Utils._error(self, f"'{inventoryName}' does not exist.")
                return

            # where
            try:
                topLevelWhere, resolvedFilterDict = Utils._handleWhere(
                    self, where, inventoryName)
            except:
                return

            validityDate = Utils._argNone('validityDate', validityDate)
            allValidityPeriods = Utils._argNone(
                'allValidityPeriods', allValidityPeriods)

            # core
            deleteId = False
            propertyDict = Utils._properties(self.scheme, inventoryName, recursive=True,
                sysProperties=includeSysProperties, maxRecursionDepth=maxRecursionDepth)
            if fields != None:
                if type(fields) != list:
                    fields = [fields]
                if 'sys_inventoryItemId' not in fields:
                    deleteId = True
                    fields += ['sys_inventoryItemId']
                _fields = Utils._queryFields(fields, propertyDict['arrayTypeFields'], arrayPageSize,
                                             filter=resolvedFilterDict, recursive=True)
            else:
                properties = Utils._propertyList(
                    propertyDict['properties'], recursive=references)
                _fields = Utils._queryFields(properties, propertyDict['arrayTypeFields'],
                                             arrayPageSize, filter=resolvedFilterDict, recursive=references)
            logger.debug(f"Fields: {_fields}")

            if len(_fields) == 0:
                Utils._error(self, f"Inventory '{inventoryName}' not found.")
                return

            order = Utils._orderItems(self, orderBy, asc)
            if order == None:
                return

            result = []
            count = 0
            stop = False
            lastId = ''

            while True:

                # Handling top (premature stop)
                if top != None:
                    loadedItems = pageSize * count
                    if top - loadedItems <= pageSize:
                        stop = True
                        pageSize = top - loadedItems

                graphQLString = f''' query getItems {{
                        {inventoryName} (
                                pageSize: {pageSize}
                                {order}
                                {allValidityPeriods}
                                {validityDate}
                                {lastId}
                                {topLevelWhere}
                                ) {{
                            {_fields}
                        }}
                    }}
                    '''

                _result = Utils._executeGraphQL(
                    self, graphQLString, correlationId)

                if _result[inventoryName]:
                    result += _result[inventoryName]
                    count += 1
                try:
                    cursor = _result[inventoryName][-1]['sys_inventoryItemId']
                    lastId = f'lastId: "{cursor}"'
                except:
                    break

                if stop == True:
                    break

            df = pd.json_normalize(result)

            nested_columns = [col for col in df.columns if "." in col]
            if nested_columns:
                base_property_names = [col.split('.')[0] for col in nested_columns]
                subset_columns = [col for col in df.columns if col not in base_property_names]
                df = df[subset_columns]
        
            if fields != None: #sorts dataframe according to given fields
                try: # for array field this does not work
                    df = df[fields]
                except:
                    pass
            if deleteId:
                try:
                    del df['sys_inventoryItemId']
                except:
                    pass

            return df

    def inventoryProperties(
        self,
        inventoryName,
        namesOnly=False
    ) -> pd.DataFrame:
        """
        Returns a DataFrame of a query of properties of an inventory.

        Parameters:
        ----------
        inventoryName : str
            The name of the inventory.
        namesOnly : bool
            If True, only property names will be returned

        Example:
        --------
        >>> inventoryProperties('appartments') 


        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            propertyFields = f'''
                name          
                ... Scalar
                isArray
                nullable
                ... Reference
                type
                propertyId
            '''

            graphQLString = f'''query Inventory {{
            inventory
                (inventoryName: "{inventoryName}")
                {{
                properties {{
                    {propertyFields}
                    }}
                }}
            }}
            fragment Scalar on IScalarProperty {{
                dataType
                }}
            fragment Reference on IReferenceProperty {{
                inventoryId
                inventoryName
            }}
            '''

            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result['inventory'] == None:
                Utils._error(self, f"Inventory '{inventoryName}' not found.")
                return

            df = pd.json_normalize(result['inventory']['properties'])

            if namesOnly == True:
                return list(df['name'])
            else:
                return df

    def propertyList(self, inventoryName, references=False, dataTypes=False, maxRecursionDepth=2):
        """
        Returns a list of properties of an inventory and its referenced inventories
        by reading out the scheme.

        Parameters:
        ----------
        inventoryName : str
            The name of the inventory.
        references : bool
            If True, properties of referenced inventories included.
        dataTypes : bool
            If True, result will be displayed as Series with properties as index and
            dataTypes as values.
        maxRecursionDepth: int = 2
            Maximum recursion level following the references to other inventories.

        Example:
        --------
        >>> propertyList('appartments') 

        """

        propertyDict = Utils._properties(self.scheme, inventoryName=inventoryName, recursive=references, maxRecursionDepth=maxRecursionDepth)

        if dataTypes == False:
            properties = Utils._propertyList(
                propertyDict['properties'], recursive=references)
        else:
            properties = pd.Series(
                Utils._propertyTypes(propertyDict['properties']))

        return properties

    def filterValues(
        self,
        inventoryName: str,
        top: int = 10000
    ) -> pd.DataFrame:
        """
        Returns a DataFrame of values that can be used in a where-string. 
        Only string data types are returned. 

        Parameters:
        ----------
        inventoryName : str
            The name of the inventory.
        top: int = None
            Uses a restricted set of items oriented at the latest entries to 
            provide a value set.

        Example:
        --------
        >>> filterValues('appartments') 

        """
        properties = self.propertyList(inventoryName, dataTypes=True)
        logger.debug(f"Properties: {properties}")

        propertyList = []
        for property, dataType in zip(properties.index, properties):
            if 'sys_' in property:
                continue
            if dataType != 'String':
                continue
            propertyList.append(property)
        logger.debug(f"PropertyList: {propertyList}")

        df = self.items(inventoryName, fields=propertyList, top=top)
        logger.debug(f"Used columns: {df.columns}")

        propertyValues = {}
        for property in propertyList:
            if 'sys_' in property:
                continue
            if len(set(df[property])) == len(df):
                continue
            propertySet = set(df[property])
            logger.debug(f"PropertySet: {propertySet}")
            try:
                propertySet.remove(None)
            except:
                pass
            propertyValues.setdefault(
                df[property].name, sorted(list(propertySet)))

        valuesDf = pd.DataFrame.from_dict(propertyValues, orient='index').T
        valuesDf.fillna(value='-', inplace=True)
        return valuesDf

    def __addItems(
        self,
        inventoryName: str,
        items: list
    ) -> str:
        """
        Adds from a list of dicts new items and returns a list
        of inventoryItemIds.

        Parameters:
        -----------
        inventoryName : str
            The name of the inventory.
        items : list
            A list with dictionaries for each item.

        Example:
        --------
        >>> items = [
                {
                'meterId': '86IEDD99',
                'dateTime': '2020-01-01T05:50:59Z'
                },
                {
                'meterId': '45IXZ52',
                'dateTime': '2020-01-07T15:41:14Z'
                }
            ]
        >>> addItems('meterData', items)
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):

            items = Utils._propertiesToString(items)
            key = f'create{inventoryName}'

            graphQLString = f'''mutation addItems {{
                {key} (input: 
                    {items}
                )
                    {{
                        {Utils.errors}           
                    inventoryItems {{
                        sys_inventoryItemId
                    }}
                }}
            }} 
            '''
            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return

            if result[key]['errors']:
                Utils._listGraphQlErrors(result, key)
            return result[key]['inventoryItems']

    def addItems(
        self,
        inventoryName:str, 
        items:list,
        chunkSize:int = 5000, 
        pause:int = 1
        ) -> str:
        """
        Adds from a list of dicts new items and returns a list
        of inventoryItemIds.

        Parameters:
        -----------
        inventoryName : str
            The name of the inventory.
        items : list
            A list with dictionaries for each item.
        chunkSize : int = 5000
            Determines the number of items which are written per chunk. Using chunks
            can be necessary to avoid overloading. Default is 5000 items per chunk.
        pause : int = 1
            Pause in seconds between each chunk upload to avoid overloading.

        Example:
        --------
        >>> items = [
                {
                'meterId': '86IEDD99',
                'dateTime': '2020-01-01T05:50:59Z'
                },
                {
                'meterId': '45IXZ52',
                'dateTime': '2020-01-07T15:41:14Z'
                }
            ]
        >>> addItems('meterData', items)        
        """
        
        correlationId = str(uuid4())
        result = []
        with logger.contextualize(correlation_id=correlationId):
            if len(items) > chunkSize:
                lenResult = 0
                for i in range(0, len(items), chunkSize):
                    result_object = self.__addItems(inventoryName, items[i : i + chunkSize])
                    if result_object != None:
                        result.extend(result_object)
                        lenResult = len(result)
                    logger.info(f"{lenResult} items of {len(items)} added. Waiting {pause} second(s) before continuing...")
                    sleep(pause)
            else:
                result = self.__addItems(inventoryName, items)
                logger.info(f"{len(result)} items added.")
        
        return result

    def __addValidityItemsToParents(
        self, 
        inventoryName:str, 
        items:list
    ) -> str:
        """
        Adds from a list of dicts items with validity periods to an existing parent items. 
        The 'sys_parentInventoryItemId' and either 'sys_validFrom' or 'sys_validFrom' are required
        system properties.

        Parameters:
        -----------
        inventoryName : str
            The name of the inventory.
        items : list
            A list with dictionaries for each item.

        Example:
        --------
        >>> items = [
                {
                'meterId': '86IEDD99',
                'userId': 'DlvK5PCm4u',
                'sys_parentInventoryItemId': 'EaM9zHA8Mi',
                'sys_validFrom': '2023-12-31T23:00:00.000Z',
                'sys_validTo': '2024-06-30T23:00:00.000Z',
                },
                {
                'meterId': '86IEDD99',
                'userId': 'DlvK5PCm8i',
                'sys_parentInventoryItemId': 'EaM9zHA8Mi',
                'sys_validFrom': '2024-07-31T23:00:00.000Z',
                'sys_validTo': '2024-09-30T23:00:00.000Z',
                }
            ]
        >>> addValidityItemsToParents('meterData', items)
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            items = Utils._propertiesToString(items)
            key = f'create{inventoryName}ValidityPeriods'

            graphQLString = f'''mutation addValidityPeriodItemsToParents {{
                {key} (input: 
                    {items}
                )
                    {{
                        {Utils.errors}           
                    inventoryItems {{
                        sys_inventoryItemId
                        sys_versionId
                    }}
                }}
            }} 
            '''
            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return

            if result[key]['errors']:
                Utils._listGraphQlErrors(result, key)
            return result[key]['inventoryItems']

    def addValidityItemsToParents(
        self, 
        inventoryName:str, 
        items:list,
        chunkSize:int = 5000, 
        pause:int = 1
        ) -> str:
        """
        Adds from a list of dicts items with validity periods to an existing parent items. 
        The 'sys_parentInventoryItemId' and either 'sys_validFrom' or 'sys_validFrom' are required
        system properties.

        Parameters:
        -----------
        inventoryName : str
            The name of the inventory.
        items : list
            A list with dictionaries for each item.
        chunkSize : int = 5000
            Determines the number of items which are written per chunk. Using chunks
            can be necessary to avoid overloading. Default is 5000 items per chunk.
        pause : int = 1
            Pause in seconds between each chunk upload to avoid overloading.

        Example:
        --------
        >>> items = [
                {
                'meterId': '86IEDD99',
                'userId': 'DlvK5PCm4u',
                'sys_parentInventoryItemId': 'EaM9zHA8Mi',
                'sys_validFrom': '2023-12-31T23:00:00.000Z',
                'sys_validTo': '2024-06-30T23:00:00.000Z',
                },
                {
                'meterId': '86IEDD99',
                'userId': 'DlvK5PCm8i',
                'sys_parentInventoryItemId': 'EaM9zHA8Mi',
                'sys_validFrom': '2024-07-31T23:00:00.000Z',
                'sys_validTo': '2024-09-30T23:00:00.000Z',
                }
            ]
        >>> addValidityItemsToParents('meterData', items)
        """

        correlationId = str(uuid4())
        result = []
        with logger.contextualize(correlation_id=correlationId):
            if len(items) > chunkSize:
                lenResult = 0
                for i in range(0, len(items), chunkSize):
                    result_object = self.__addValidityItemsToParents(inventoryName, items[i : i + chunkSize])
                    if result_object != None:
                        result.extend(result_object)
                        lenResult = len(result)
                    logger.info(f"{lenResult} validity items of {len(items)} added. Waiting {pause} second(s) before continuing...")
                    sleep(pause)
            else:
                result = self.__addValidityItemsToParents(inventoryName, items)
                logger.info(f"{len(result)} validity items added.")
        
        return result

    def updateItems(
        self,
        inventoryName: str,
        items: list
    ) -> None:
        """
        Updates from a list of dicts existing items. The 'sys_inventoryItemId'
        must be passed to each item.

        Parameters:
        -----------
        inventoryName : str
            The name of the inventory.
        items : list
            A list with dictionaries for each item.

        Example:
        --------
        >>> items = [
                {
                'sys_inventoryItemId':'118312438662692864',
                'meterId': '86IEDD99',
                'dateTime': '2020-01-01T05:50:59Z'
                },
                {
                'sys_inventoryItemId':'118312438662692864',
                'meterId': '45IXZ52',
                'dateTime': '2020-01-07T15:41:14Z'
                }
            ]
        >>> updateItems('meterData', items)
        """
        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            items = Utils._propertiesToString(items)
            key = f'update{inventoryName}'

            graphQLString = f'''mutation updateItems {{
                {key} (input: 
                    {items}
                )
                    {{
                        {Utils.errors}           
                        inventoryItems {{
                            sys_inventoryItemId
                    }}
                }}
            }}
            '''
            logger.trace(graphQLString)
            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return

            if result[key]['errors']:
                Utils._listGraphQlErrors(result, key)

            logger.info(
                f"{len(result[key]['inventoryItems'])} item(s) updated.")
            return

    def updateDataFrameItems(
        self,
        inventoryName: str,
        dataFrame: pd.DataFrame,
        columns: list = None
    ) -> None:

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            def convertDfToDict(df: pd.DataFrame, timeSeries: bool, columns: list = None) -> dict:

                def createDict(df: pd.DataFrame) -> dict:
                    items = []
                    for _, row in df.iterrows():
                        item = {}
                        for header, value in zip(df.columns, row):
                            if not pd.isna(value):
                                item.setdefault(header, value)
                        items.append(item)
                    return items

                if timeSeries == True:
                    tsProperties = ['resolution', 'unit']
                    if columns == None:
                        columns = [
                            col for col in df.columns if col not in tsProperties]
                    else:
                        columns = [
                            col for col in columns if col not in tsProperties]

                if columns == None:
                    return createDict(df)
                else:
                    if 'sys_inventoryItemId' not in columns:
                        columns.append('sys_inventoryItemId')
                    return df[columns].to_dict('records', )

            if not 'sys_inventoryItemId' in dataFrame.columns:
                logger.error(
                    "Missing column 'sys_inventoryItemId'. Items cannot be updated without this information.")
                return

            if self.structure[inventoryName]['variant'] == None:
                items = convertDfToDict(dataFrame, False, columns=columns)
            elif self.structure[inventoryName]['variant']['name'] not in ['TimeSeries', 'TimeSeriesGroup']:
                items = convertDfToDict(dataFrame, False, columns=columns)
            else:
                items = convertDfToDict(dataFrame, True, columns=columns)

            logger.debug(f"Items to write {items[:3]}...")

            self.updateItems(inventoryName, items)
            return

    def createInventory(
        self,
        name: str,
        properties: list,
        variant: str = None,
        propertyUniqueness: dict = None,
        historyEnabled: bool = False,
        hasValitityPeriods: bool = False,
        isDomainUserType: bool = False
    ) -> str:
        """
        Creates a new inventory. After creation, access rights must be set to add items.

        Parameters:
        ----------
        name : str
            Name of the new inventory (only alphanumeric characters allowed, 
            may not begin with a number)
        properties : list
            A list of dicts with the following mandatory keys: 
                name: str
                dataType: enum (STRING, BOOLEAN, DECIMAL, INT, LONG, DATE_TIME, 
                DATE_TIME_OFFSET)
            Optional keys:
                isArray: bool (Default = False)
                nullable: bool (Default = True)
                isReference: bool (Default = False)
                inventoryId: str (mandatory if hasReference = True)
        variant : str
            The inventory variant.
        propertyUniqueness : list
            A list of properties that should be unique in its combination. 
        historyEnabled : bool
            If True, changes in properties will be recorded in item history.
        hasValidityPeriods : bool
            If true, a validity period can be added to the item.    

        Example:
        --------
        >>> propertyDefinitions = [
            {
                'name': 'street',
                'dataType': 'STRING',
                'nullable': False,
            },
            {
                'name': 'postCode',
                'dataType': 'STRING',
                'nullable': False,
            },
            ]
            uniqueness = [{'uniqueKey': 'address', 'properties': ['street', 'postCode']}
        >>> createInventory('appartment', 'propertyDefinitions', propertyUniqueness=uniqueness) 
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            _properties = Utils._propertiesToString(properties)

            if variant != None:
                _variantId = f'{Utils._getVariantId(self.variants(), variant)}'
                logger.debug(f"Found variantId: {_variantId}")
                if type(_variantId) != str:
                    Utils._error(self, f"Variant name '{name}' not found")
                    return

                _variant = f'variantId: "{_variantId}"'
            else:
                _variant = ''

            if propertyUniqueness != None:
                _propertyUniqueness = 'propertyUniqueness: ' + \
                    Utils._uniquenessToString(propertyUniqueness)
                logger.debug(_propertyUniqueness)
            else:
                _propertyUniqueness = ''

            _history = 'true' if historyEnabled != False else 'false'
            _validityPeriods = 'true' if hasValitityPeriods != False else 'false'
            _domainUser = 'true' if isDomainUserType != False else 'false'

            graphQLString = f'''
            mutation createInventory {{
                createInventory (input: {{
                    name: "{name}"        
                    properties: {_properties}
                    {_variant}
                    historyEnabled: {_history}
                    hasValidityPeriods: {_validityPeriods}
                    isDomainUserType: {_domainUser}
                    {_propertyUniqueness}
                    }})
                    {{
                    inventory {{
                        inventoryId
                        }}
                    {Utils.errors}           
                }}
            }}
            '''
            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return

            if result['createInventory']['errors']:
                Utils._listGraphQlErrors(result, 'createInventory')
                return

            logger.info(f"New inventory {name} created.")

            if variant != None:
                self.TimeSeries.refreshSchema()

            return result['createInventory']['inventory']['inventoryId']

    def deleteInventories(
        self,
        inventoryNames: list,
        deleteWithData: bool = False,
        force: bool = False
    ) -> None:
        """ 
        Deletes one or more inventories with the possibility to delete containing data.

        Parameters:
        -----------
        inventoryNames : list
            A list of inventory names that should be deleted.
        deleteWithData : bool = False
            If True, containing data will be deleted.
        force : bool
            Use True to ignore confirmation.

        Example:
        ---------
        >>> deleteInventories(['meterData'], deleteWithData=True, force=True)
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            inventoryIds = []
            errorNames = []
            for name in inventoryNames:
                if deleteWithData and not self._isInventoryOfValidVariant(name):
                    Utils._error(self, f"TimeSeries inventory {name} cannot be deleted with option 'deleteWithData'.")
                    continue
                try:
                    inventoryIds.append(self.structure[name]['inventoryId'])
                except:
                    errorNames.append(name)

            if errorNames:
                Utils._error(self, f"Unknown inventory names '{errorNames}'.")
                return
            if len(inventoryIds) == 0:
                logger.info(f"No inventories to delete.")
                return

            _inventoryIds = Utils._graphQLList(inventoryIds)

            if force == False:
                confirm = input(f"Press 'y' to delete '{inventoryNames}': ")

            key = 'deleteInventory'
            graphQLString = f'''mutation deleteInventories {{
                {key} (input:
                    {{
                        inventoryIds: {_inventoryIds}
                        ignoreData: {Utils._toGraphQL(deleteWithData)}
                    }}
                )
                    {{
                        {Utils.errors}           
                    }}
                }}
                '''

            if force == True:
                confirm = 'y'
            if confirm == 'y':
                result = Utils._executeGraphQL(
                    self, graphQLString, correlationId)
                if result == None:
                    return
                if result[key]['errors']:
                    Utils._listGraphQlErrors(result, key)
                    return
                else:
                    logger.info(f"Inventories '{inventoryNames}' deleted.")
            else:
                return

    def variants(self) -> pd.DataFrame:
        """
            Returns a dataframe of available variants.
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            key = 'variants'
            graphQLString = f'''query getVariants {{
            {key} {{
                name
                variantId
                }}
            }}
            '''
            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return

            df = pd.json_normalize(result[key])
            return df

    def deleteVariant(
        self,
        variantId: str,
        force: bool = False
    ) -> None:
        """Deletes a variant

        Parameters
        -----------
        variantId : str
            The id of the variant.
        force : bool
            Use True to ignore confirmation.
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):

            if force == False:
                confirm = input(
                    f"Press 'y' to delete  variant with Id {variantId}: ")

            key = 'deleteVariant'
            graphQLString = f'''mutation deleteVariant {{
                {key}(input: {{ variantId: "{variantId}" }}) {{
                    errors {{
                    message
                    code
                    }}
                }}
            }}
            '''

            if force == True:
                confirm = 'y'
            if confirm == 'y':
                result = Utils._executeGraphQL(
                    self, graphQLString, correlationId)
                if result == None:
                    return
            else:
                return

            if result[key]['errors'] != None:
                Utils._listGraphQlErrors(result, key)
            else:
                logger.info(f"Variant deleted.")

    def deleteItems(
        self,
        inventoryName: str,
        inventoryItemIds: list = None,
        where: str = None,
        force: bool = False,
        pageSize: int = 500
    ) -> None:
        """
        Deletes inventory items from a list of inventoryItemIds or by where-criteria. 

        Parameters:
        -----------
        inventoryName: str
            The name of the inventory.
        inventoryItemIds: list = None
            A list of inventoryItemIds that should be deleted.
        where: str = None
            Filter criteria to select items that should be deleted.
        force: bool = False
            Use True to ignore confirmation.
        pageSize: int = 500
            Only a limited amount of items can be deleted at once. 500 is default, however, 
            if this size is too, big, choose a lower pageSize.

        Examples:
        ---------
        >>> deleteItems('meterData', where='changeDate gt "2020-12-01"', force=True, pageSize=100)
        >>> deleteItems('meterData', inventoryItemIds=['El5JrMG2xk'])
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            if not self._isInventoryOfValidVariant(inventoryName):
                Utils._error(self, f"Items of a TimeSeries inventory cannot be deleted. Use TimeSeries.deleteItems() instead.")
                return
            def delete(ids, n, m):

                _ids = '['
                for id in ids[n:m]:
                    _ids += f'''{{sys_inventoryItemId: "{id}"}}\n'''
                _ids += ']'

                key = f'delete{inventoryName}'
                graphQLString = f'''
                    mutation deleteItems {{
                        {key} ( input: 
                            {_ids}
                        )
                        {{
                            {Utils.errors}           
                        }}
                    }}
                    '''

                result = Utils._executeGraphQL(
                    self, graphQLString, correlationId)

                if result[key]['errors'] != None:
                    logger.error(Utils._listGraphQlErrors(result, key))
                    return

            if inventoryItemIds == None and where == None:
                Utils._error(
                    self, f"No list of items and no where-criteria were provided.")
                return

            if inventoryItemIds != None and where != None:
                logger.warning(
                    f"List of items and where-criteria has been provided. Item list is used.")

            if where != None:
                _result = self.items(inventoryName, fields=[
                                     'sys_inventoryItemId'], where=where)
                if _result.empty:
                    logger.info(f"No results found for provided filter.")
                    return
                ids = list(_result['sys_inventoryItemId'])
            if inventoryItemIds != None:
                _result = self.items(inventoryName, fields=[
                                     'sys_inventoryItemId'], where=f'sys_inventoryItemId in {inventoryItemIds}')
                if _result.empty:
                    logger.info(f"Provided id(s) could not be found.")
                    return
                ids = list(_result['sys_inventoryItemId'])
                diff = set(inventoryItemIds).difference(set(ids))
                if diff:
                    Utils._error(
                        self, f"The following item id's are not in the inventory: {ids}")
                    return

            logger.debug(f"GraphQL Ids: {ids}")

            if force == False:
                confirm = input(f"Press 'y' to delete  {len(ids)} items: ")

            if force == True:
                confirm = 'y'
            if confirm == 'y':
                n = 0
                m = n + pageSize

                while True:
                    delete(ids, n, m)
                    n += pageSize
                    m += pageSize

                    if len(ids) - m < pageSize:
                        delete(ids, n, len(ids))
                        break
            else:
                return

            logger.info(f"{len(ids)} items deleted.")
            return

    def clearInventory(
        self,
        inventoryName: str,
        force: bool = False,
        pageSize: int = 500
    ) -> None:
        """
        Deletes all items from the inventory

        Parameters
        -----------
        inventoryName : str
            The name of the inventory.
        force : bool
            Use True to ignore confirmation.
        pageSize : str = 500
            The number of items to be deleted in a chunk (the maximum number 
            of items that can be deleted in one mutation is restricted).

        Example:
        ---------
        >>> clearInventory('meterData', force=True)
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            if not self._isInventoryOfValidVariant(inventoryName):
                Utils._error(self, f"Items of a TimeSeries inventory cannot be deleted. Use TimeSeries.deleteItems() instead.")
                return
            if force == False:
                confirm = input(
                    f"Press 'y' to delete all items in inventory {inventoryName}: ")
            if force == True:
                confirm = 'y'
            if confirm != 'y':
                return

            count = 0
            lastId = ''

            while True:
                graphQLString = f''' query getItems {{
                        {inventoryName} (
                                pageSize: {pageSize} 
                                {lastId}
                                ) {{
                            sys_inventoryItemId
                        }}
                    }}
                    '''

                _result = Utils._executeGraphQL(
                    self, graphQLString, correlationId)
                ids = [item['sys_inventoryItemId']
                       for item in _result[inventoryName]]
                count += len(ids)

                _ids = ''
                for id in ids:
                    _ids += f'{{sys_inventoryItemId: "{id}"}}\n'

                try:
                    cursor = _result[inventoryName][-1]['sys_inventoryItemId']
                    lastId = f'lastId: "{cursor}"'

                    key = f'delete{inventoryName}'
                    graphQLString = f'''
                        mutation deleteItems {{
                            {key} ( input: 
                                [{_ids}]
                            )
                            {{
                                {Utils.errors}           
                            }}
                        }}
                    '''
                    result = Utils._executeGraphQL(
                        self, graphQLString, correlationId)
                except:
                    break

            if count == 0:
                logger.info(f"Inventory is empty. No items were deleted.")
                return
            if result[key]['errors'] != None:
                Utils._listGraphQlErrors(result, key)
            else:
                logger.info(f"{count} items deleted.")

    def updateVariant(
        self,
        variantName,
        newName=None,
        icon=None
    ) -> None:
        """Updates a variant"""

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            _variantId = Utils._getVariantId(self.variants(), variantName)
            logger.debug(f"Found variantId: {_variantId}")

            if newName != None:
                _name = f'name: "{newName}"'
            else:
                _name = 'null'

            if icon != None:
                _icon = f'icon: "{icon}"'
            else:
                _icon = 'null'

            key = 'updateVariant'
            graphQLString = f'''mutation updateVariant {{
                {key} (input:{{
                    variantId: "{_variantId}"
                    {_name}
                    {_icon}
                    }})
                    {{
                        {Utils.errors}           
                    }}
                }}
                '''

            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return

            return

    def updateArrayProperty(
        self,
        inventoryName: str,
        inventoryItemId: str,
        arrayProperty: str,
        operation: str,
        arrayItems: list = None,
        cascadeDelete: bool = False
    ) -> None:
        """
        Updates a single array property of a single inventoryItemId. Arrays with and without 
        references are supported.

        Parameters:
        -----------
        inventoryName: str
            The name of the inventory where the item is located.
        inventoryItemId: str
            The sys_inventoryItemId of the item.
        arrayProperty: str
            The name of the property whose array items are to be updated.
        operation: str
            The update operation to be performed. Options:
            insert: inserts a list of array items.
            removeById: removes array items by a list of given ids.
            removeByIndex: removes array items by a list of given indices.
            removeAll: removes all array items
        arrayItems: list = None
            A list of indices or item Ids
        cascadeDelete: bool = False
            If array items are refencences, True will delete also the reference items. 

        Examples:
        ---------
        >>> client.updateArrayProperty('meterData', 'A5N45hOOmm',
                'timeSeries', action='insert', arrayItems=['A5FdSjehbE'])
        >>> client.updateArrayProperty('meterData', 'A5N45hOOmm',
                'timeSeries', action='removeByIndex', arrayItems=[0,1], cascadeDelete=True)
        """
        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):

            operations = ['insert', 'removeById', 'removeByIndex', 'removeAll']
            if operation not in operations:
                Utils._error(
                    self, f"Action '{operation}' allowed. Possible update operations: {operations}.")
                return

            if operation == 'removeAll':
                try:
                    propertyDf = TechStack.inventoryProperties(
                        self, inventoryName)
                    arrayProps = propertyDf[propertyDf['name']
                                            == arrayProperty]
                    logger.debug(f"Properties of array: {arrayProps}")
                    if arrayProps['type'].item() == 'reference':
                        # arrDf = TechStack.items(self, inventoryName,
                        #     where=f'sys_inventoryItemId eq "{inventoryItemId}"', fields=[f'{arrayProperty}.sys_inventoryItemId'],
                        #     arrayPropertiesToPage=[arrayProperty], arrayPageSize=5000)
                        arrDf = TechStack.items(self, inventoryName,
                                                where=f'sys_inventoryItemId eq "{inventoryItemId}"', fields=[f'{arrayProperty}.sys_inventoryItemId'],
                                                )
                    else:
                        arrDf = TechStack.items(self, inventoryName, fields=[{arrayProperty}],
                                                where=f'sys_inventoryItemId eq "{inventoryItemId}"', arrayPageSize=5000)
                except Exception as err:
                    Utils._error(self, err)
                    return

                countArray = len(arrDf[arrayProperty].item())
                arrayItems = [num for num in range(countArray)]
                logger.debug(f"Array Items: {arrayItems}")
            _arrayItems = Utils._arrayItemsToString(
                arrayItems, operation, cascadeDelete)
            logger.debug(f"Array Items as String: {_arrayItems}")

            key = f'update{inventoryName}ArrayProperties'
            graphQLString = f'''mutation updateArray {{
                {key} (
                    input: {{
                    sys_inventoryItemId: "{inventoryItemId}"
                    {arrayProperty}: {{
                        {_arrayItems}
                        }}
                    }}
                ) 
                {{
                    {Utils.errors}           
                }}
            }}'''

            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return

            if result[key]['errors'] != None:
                Utils._listGraphQlErrors(result, key)
            else:
                logger.info(
                    f"Array property {arrayProperty} for item {inventoryItemId} updated.")

    def addInventoryProperties(
        self,
        inventoryName: str,
        properties: list
    ) -> None:
        """
        Adds one or more inventory properties to an existing inventory.

        Parameters:
        ----------
        inventoryName: str
            Name of inventory
        properties: list
            A list of dicts with the following mandatory keys: 
                name: str
                dataType: enum (STRING, BOOLEAN, DECIMAL, INT, LONG, DATE_TIME, 
                DATE_TIME_OFFSET)
            Optional keys:
                isArray: bool (Default = False)
                nullable: bool (Default = True)
                isReference: bool (Default = False)
                inventoryId: str (mandatory if hasReference = True)
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            key = 'addProperty'
            inventory = self.inventories(where=f'name eq "{inventoryName}"')

            if inventory.empty:
                Utils._error(self, f"Unknown inventory '{inventoryName}'.")
                return

            inventoryId = inventory.loc[0, 'inventoryId']
            _properties = Utils._propertiesToString(properties)

            graphQLString = f'''
            mutation {key} {{
            {key} (input: {{
                inventoryId: "{inventoryId}"	
                properties: {_properties}
                }}) 
                {{
                    {Utils.errors}           
                }}
            }}
            '''
            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return

            if result[key]['errors']:
                Utils._listGraphQlErrors(result, key)
                return
            else:
                self.updateClient()
                logger.info(f"New property(ies) added.")

    def updateDisplayValue(
        self,
        inventoryName: str,
        displayValue: str
    ) -> None:

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            inventoryId = self.structure[inventoryName]['inventoryId']

            key = 'updateInventory'
            graphQLString = f'''mutation updateDisplayValue {{
                {key} (input: {{
                    inventoryId: "{inventoryId}"
                    displayValue: "{displayValue}"
                    }}) {{
                    {Utils.errors}
                }}
            }}
            '''
            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return

            if result[key]['errors']:
                Utils._listGraphQlErrors(result, key)
                return
            else:
                self.updateClient()
                logger.info(f"Display value updated.")
                return

    def updateInventoryName(
        self,
        inventoryName: str,
        newName: str
    ) -> None:

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            inventoryId = self.structure[inventoryName]['inventoryId']
            key = 'updateInventory'

            graphQLString = f'''mutation updateInventoryName {{
                {key} (input: {{
                    name: "{newName}"
                    inventoryId: "{inventoryId}"
                }}) {{
                    {Utils.errors}
                    }}
                }}
            '''
            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return
            if result[key]['errors']:
                Utils._listGraphQlErrors(result, key)
                return
            else:
                logger.info(f"Inventory name updated.")
                self.updateClient()
                return

    def removeProperties(
        self,
        inventoryName: str,
        properties: list
    ) -> None:
        """
        Removes a list of properties given as property names. Properties can 
        only be removed if they have no content or are of type 'reference'.

        Parameters:
        -----------
        inventoryName: str
            The name of the inventory where the item is located.
        properties: list
            A list of property names.
        """

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            if type(properties) != list:
                properties = [properties]

            propertyList = '['
            for argProperty in properties:
                try:
                    propertyId = self.structure[inventoryName]['properties'][argProperty]['propertyId']
                    propertyList += f'"{propertyId}",'
                except:
                    logger.warning(
                        f"Property '{argProperty}' not found in inventory '{inventoryName}'.")
            propertyList += ']'

            inventoryId = self.structure[inventoryName]['inventoryId']

            key = 'removeProperty'
            graphQLString = f'''mutation removeProperty {{
                {key} (input: {{
                        inventoryId: "{inventoryId}"
                        propertyIds: {propertyList}
                }}) {{
                    {Utils.errors}
                    }}
                }}
            '''
            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return
            if result[key]['errors']:
                Utils._listGraphQlErrors(result, key)
                return
            else:
                self.updateClient()
                logger.info(f"Properties have been removed.")
                return

    def updateProperty(
        self,
        inventoryName: str,
        propertyName: str,
        newPropertyName: str = None,
        nullable: bool = None
    ) -> None:

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            inventoryId = self.structure[inventoryName]['inventoryId']
            newPropertyName = Utils._argNone('name', newPropertyName)
            nullable = Utils._argNone('nullable', nullable)

            try:
                propertyId = self.structure[inventoryName]['properties'][propertyName]['propertyId']
            except:
                Utils._error(
                    self, f"Property '{propertyName}' not found in inventory '{inventoryName}'.")
                return

            key = 'updateInventory'
            graphQLString = f'''mutation removeProperty {{
                {key} (input: {{
                        inventoryId: "{inventoryId}"
                        properties: [{{
                            propertyId: "{propertyId}"
                            {newPropertyName}
                            {nullable}
                        }}]
                }}) {{
                    {Utils.errors}
                    }}
                }}
            '''
            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return
            if result[key]['errors']:
                Utils._listGraphQlErrors(result, key)
                return
            else:
                self.updateClient()
                logger.info(f"Property updated.")
                return

    def resync(self) -> None:
        """Resynchronizes read databases"""

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            key = 'reSyncReadDatabase'
            graphQLString = f'''mutation resync{{
                {key}
                }}
            '''
            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return
            self.updateClient()

            return

    def defaultDataFrame(
        self,
        maxRows,
        maxColumns
    ) -> None:
        """Sets default sizes for a DataFrame for the current session"""
        pd.options.display.max_rows = maxRows
        pd.options.display.max_columns = maxColumns
        return

    def _convertId(
        self,
        sys_inventoryItemId: str
    ) -> str:
        """Convers a sys_inventoryItemId into a HAKOM name"""

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            graphQLString = f'''mutation convert{{
                convertId(id: "{sys_inventoryItemId}")
                }}
            '''
            result = Utils._executeGraphQL(self, graphQLString, correlationId)
            if result == None:
                return

            return result['convertId']

    def _isInventoryOfValidVariant(
        self,
        inventoryName: str,
        variantName: str = None
    ) -> bool:
        """Checks if an inventory is of a valid variant"""

        correlationId = str(uuid4())
        with logger.contextualize(correlation_id=correlationId):
            inventory = self.inventories(
                where=f'name eq "{inventoryName}"', fields=['variant.name'])
            if inventory.empty:
                Utils._error(self, f"Unknown inventory '{inventoryName}'.")
                return

            # in case of inventory without variant there is a column 'variant' with None
            if variantName == None:
                return 'variant' in inventory
            else:
                return not 'variant' in inventory and inventory.iloc[0]['variant.name'] == variantName


import pandas
import pytz
import csv

from re import match
from loguru import logger
from tzlocal import get_localzone
from datetime import datetime
from pathlib import Path


class UtilsProperty():
    def _getDataTypes(properties: dict) -> dict:
        """
        Sorts the property types and returns a dictionary of those properties 
        that are not of type STRING.
        """

        logger.debug(properties)
        propertyTypes = {'Boolean': [], 'Numeric': [],
                         'DateTime': [], 'String': [], 'Unknown': []}
        if 'dataType' not in properties[0]:
            return propertyTypes

        for property in properties:
            if property['dataType'] in ['DATE_TIME', 'DATE_TIME_OFFSET']:
                propertyTypes['DateTime'].append(property['name'])
            elif property['dataType'] == 'BOOLEAN':
                propertyTypes['Boolean'].append(property['name'])
            elif property['dataType'] in ['INT', 'LONG', 'DECIMAL']:
                propertyTypes['Numeric'].append(property['name'])

            elif property['dataType'] == 'STRING':
                propertyTypes['String'].append(property['name'])
            else:
                propertyTypes['Unknown'].append(property['name'])
        return propertyTypes

    def _getArrays(properties: dict) -> list:
        """ Lists properties that are arrays (isArray) """

        p = [property['name']
             for property in properties if property['isArray'] == True]
        return p

    def _getNullables(properties: dict) -> list:
        """ Lists properties that are nullable """

        p = [property['name']
             for property in properties if property['nullable'] == True]
        return p

    def _getReferences(properties: dict) -> list:
        """ Lists properties that are nullable """

        p = [property['name']
             for property in properties if property['type'] == 'reference']
        return p

    def _transformBool(value: str):
        trueValues = ['ja', 'Ja', 'yes', 'Yes', 'True', 'true', 1]
        falseValues = ['nein', 'Nein', 'no', 'No', 'false', 'False', 0]
        if value in trueValues:
            return True
        elif value in falseValues:
            return False
        return None

    def _isInt(n: str):
        return bool(match(r'-?\d+$', n))

    def _isFloat(n: str) -> bool:
        return bool(match(r'-?\d+(\.\d+)$', n))

    def _transformNumeric(value: str):
        if UtilsProperty._isInt(value):
            return int(value)
        elif UtilsProperty._isFloat(value):
            return float(value)
        else:
            logger.warning(f"'{value}' is not of numeric data type!")

    def _checkIsList(pythonList: str) -> bool:
        if pythonList[0] != '[':
            return False
        elif pythonList[-1] != ']':
            return False
        else:
            return True

    def _transformList(pythonList: str, propertyType: str) -> list:
        """ Transforms a String into a Python list"""
        if not UtilsProperty._checkIsList(pythonList):
            # logger.warning(f"Value '{pythonList}' is not array type, try to transform...")
            pythonList = '[' + pythonList + ']'

        pythonList = pythonList[1:-1].split(',')
        pythonList = [element.strip() for element in pythonList]

        _pythonList = []
        for element in pythonList:
            if propertyType == 'Numeric':
                _pythonList.append(UtilsProperty._transformNumeric(element))
            elif propertyType == 'Boolean':
                _pythonList.append(UtilsProperty._transformBool(element))
            else:
                _pythonList.append(str(element))

        return _pythonList


class FileUtils(UtilsProperty):

    def _createItems(
            content: list,
            dataType: dict,
            isArray: list,
            nullable: list,
            referenceMapping: dict = None,
            type: str = None,
            uniqueProperty: str = None,
            parentItemMapping: dict = None):
        """
        Creates the dictionary that is imported by 'addItems()'

        Parameters:
        ----------
        content: list
            The CSV content as list
        dataType:dict
            Property data types returned by UtilsProperty._getPropertyTypes()
        isArray: list
            Provides information, if the column is of array type. Returned by 
            UtilsProperty._getArrays()
        nullable: list
            Provides information, if the column is nullable or not. Returned by
            UtilsProperty._getNullables()
        referenceMapping: dict = None
            Provides mapping of unique properties with sys_inventoryItemIds. Returned by
            FileUtils._createReferenceMapping()
        type: str = None
            If type is None, a 'basic' item is assumed. Another valid type is 'timeSeries'.
        """

        def _createItem(header):
            if i == 0:
                return
            if len(row) < 1:
                return
            else:
                if type == None:
                    item = {}
                if type == 'timeSeries':
                    item = {'resolution': {}}
                for j, value in enumerate(row):

                    if not value:
                        if header[j] in nullable:
                            continue
                        elif header[j] not in nullable and len(value) > 0:
                            logger.warning(
                                f"Line {i}, field '{header[j]}': Value missing for non nullable property.")
                            return
                        else:
                            pass
                    else:
                        if parentItemMapping != None:
                            if header[j] == uniqueProperty:
                                if parentItemMapping[value] != None:
                                    item.setdefault(
                                        'sys_parentInventoryItemId', parentItemMapping[value])
                        if referenceMapping != None and header[j] in referenceMapping:
                            parentHeader = header[j].split('.')[0]
                            if parentHeader in isArray:
                                nameList = UtilsProperty._transformList(
                                    value, 'String')
                                try:
                                    idList = [referenceMapping[header[j]][item]
                                              for item in nameList]
                                except KeyError as key:
                                    logger.warning(
                                        f"Line {i}, field '{header[j]}': {key} is not a valid reference.")
                                    return
                                item.setdefault(parentHeader, idList)
                            else:
                                try:
                                    item.setdefault(
                                        parentHeader, referenceMapping[header[j]][value])
                                except KeyError as key:
                                    logger.warning(
                                        f"Line {i}, field '{header[j]}': {key} is not a valid reference.")
                                    return
                        else:
                            if header[j] == 'sys_inventoryItemId':
                                if uniqueProperty == 'sys_inventoryItemId':
                                    continue
                            if header[j] in dataType['Boolean']:
                                if header[j] in isArray:
                                    item.setdefault(
                                        header[j], UtilsProperty._transformList(value, 'Boolean'))
                                else:
                                    item.setdefault(
                                        header[j], UtilsProperty._transformBool(value))
                            elif header[j] in dataType['Numeric']:
                                if header[j] in isArray:
                                    item.setdefault(
                                        header[j], UtilsProperty._transformList(value, 'Numeric'))
                                else:
                                    item.setdefault(
                                        header[j], UtilsProperty._transformNumeric(value))
                            elif header[j] in dataType['DateTime']:
                                if header[j] in isArray:
                                    item.setdefault(
                                        header[j], UtilsProperty._transformList(value, 'DateTime'))
                                else:
                                    item.setdefault(header[j], value)
                            else:
                                if type == 'timeSeries' and header[j] == 'timeUnit':
                                    item['resolution'].setdefault(
                                        'timeUnit', value)
                                elif type == 'timeSeries' and header[j] == 'factor':
                                    item['resolution'].setdefault(
                                        'factor', value)
                                elif header[j] in isArray:
                                    item.setdefault(
                                        header[j], UtilsProperty._transformList(value, 'String'))
                                else:
                                    item.setdefault(header[j], value)
                return item

        itemList = []
        validityItemList = []
        header = content[0]
        for i, row in enumerate(content):
            item = _createItem(header)
            if item != None:
                if 'sys_parentInventoryItemId' in item:
                    validityItemList.append(item)
                else:
                    itemList.append(item)
        if all((uniqueProperty, parentItemMapping)) == False:
            return itemList
        else:
            return (validityItemList, itemList)

    def _createInstanceItems(content: list, dataType: dict, isArray: list, nullable: list, idMapping: dict = None,
                             transpose: bool = False) -> list:

        if transpose == True:
            content = list(map(list, zip(*content)))

        for column in content:
            if column[0] in ['unit', 'timeUnit', 'factor']:
                del column

        itemList = []
        header = content[0]
        for i, row in enumerate(content):
            itemOkay = True  # is changed to false, if certain criteria are not met -> warning message, next item
            if i == 0:
                continue
            if len(row) < 1:
                continue
            else:
                if idMapping != None:
                    try:
                        groupInventoryItemId = idMapping[content[i][0]]
                    except:
                        logger.warning(
                            f"ImportKeyProperty '{content[i][0]}' not found.")
                        continue
                else:
                    groupInventoryItemId = content[i][0]
                item = {'sys_groupInventoryItemId': groupInventoryItemId}
                for j, field in enumerate(row):
                    if j == 0:
                        continue

                    if not field:
                        if header[j] in nullable:
                            continue
                        elif header[j] not in nullable and len(field) > 0:
                            logger.warning(
                                f"Value missing for non nullable {header[j]}. Item from line {i} not imported.")
                            itemOkay = False
                            break
                        else:
                            pass
                    else:
                        if header[j] in dataType['Boolean']:
                            if header[j] in isArray:
                                item.setdefault(
                                    header[j], UtilsProperty._transformList(field, 'Boolean'))
                            else:
                                item.setdefault(
                                    header[j], UtilsProperty._transformBool(field))
                        elif header[j] in dataType['Numeric']:
                            if header[j] in isArray:
                                item.setdefault(
                                    header[j], UtilsProperty._transformList(field, 'Numeric'))
                            else:
                                item.setdefault(
                                    header[j], UtilsProperty._transformNumeric(field))
                        elif header[j] in dataType['DateTime']:
                            if header[j] in isArray:
                                item.setdefault(
                                    header[j], UtilsProperty._transformList(field, 'DateTime'))
                            else:
                                # item.setdefault(header[j], UtilsProperty._convertTimestamp(field, timeZone))
                                item.setdefault(header[j], field)
                        else:
                            if header[j] in isArray:
                                item.setdefault(
                                    header[j], UtilsProperty._transformList(field, 'String'))
                            else:
                                item.setdefault(header[j], field)
                if itemOkay and len(item) > 0:
                    itemList.append(item)

        return itemList

    def _createTimeSeriesItems(content: list, dataType: dict, isArray: list, nullable: list):
        itemList = []
        header = content[0]
        for i, row in enumerate(content):
            itemOkay = True  # is changed to false, if certain criteria are not met -> warning message, next item
            if i == 0:
                continue
            if len(row) < 1:
                continue
            else:
                item = {'resolution': {}}
                for j, field in enumerate(row):
                    if not field:
                        if header[j] in nullable:
                            continue
                        elif header[j] not in nullable and len(field) > 0:
                            logger.warning(
                                f"Value missing for non nullable {header[j]}. Item from line {i} not imported.")
                            itemOkay = False
                            break
                        else:
                            pass
                    else:
                        if header[j] in dataType['Boolean']:
                            if header[j] in isArray:
                                item.setdefault(
                                    header[j], UtilsProperty._transformList(field, 'Boolean'))
                            else:
                                item.setdefault(
                                    header[j], UtilsProperty._transformBool(field))
                        elif header[j] in dataType['Numeric']:
                            if header[j] in isArray:
                                item.setdefault(
                                    header[j], UtilsProperty._transformList(field, 'Numeric'))
                            else:
                                item.setdefault(
                                    header[j], UtilsProperty._transformNumeric(field))
                        elif header[j] in dataType['DateTime']:
                            if header[j] in isArray:
                                item.setdefault(
                                    header[j], UtilsProperty._transformList(field, 'DateTime'))
                            else:
                                # item.setdefault(header[j], UtilsProperty._convertTimestamp(field, timeZone))
                                item.setdefault(header[j], field)
                        else:
                            if header[j] == 'timeUnit':
                                item['resolution'].setdefault(
                                    'timeUnit', field)
                            elif header[j] == 'factor':
                                item['resolution'].setdefault('factor', field)
                            else:
                                if header[j] in isArray:
                                    item.setdefault(
                                        header[j], UtilsProperty._transformList(field, 'String'))
                                else:
                                    item.setdefault(header[j], field)
                if itemOkay and len(item) > 0:
                    itemList.append(item)

        return itemList

    def _comparePropertiesBasic(properties: list, header: list) -> None:
        """ Compares header with requested properties from a basic inventory. """

        properties = set(properties)
        header = set(header)
        return header.difference(properties)

    def _comparePropertiesTimeSeries(properties: list, header: list) -> None:
        """ Compares header with requested properties from a time series inventory. """

        properties = set(properties)
        header = set(header)
        tsProperties = set(['timeUnit', 'factor', 'unit'])
        header = header - tsProperties
        return header.difference(properties)

    def _analyzeProperties(inventoryName: str, properties: pandas.DataFrame) -> tuple:
        """ Analyzes inventory properties and returns dictionaries to each property attribute """

        # properties = properties[['name', 'dataType', 'isArray', 'nullable']].to_dict('records')
        properties = properties.to_dict('records')
        dataType = UtilsProperty._getDataTypes(properties)
        isArray = UtilsProperty._getArrays(properties)
        nullable = UtilsProperty._getNullables(properties)
        isReference = UtilsProperty._getReferences(properties)

        logger.debug(f'Data types: {dataType}')
        logger.debug(f'Array properties: {isArray}')
        logger.debug(f'Nullable properties: {nullable}')
        logger.debug(f'Reference properties: {isReference}')

        return dataType, isArray, nullable, isReference

    def _convertTimestamp(timestamp, timeZone, dateTimeFormat, timeDelta):

        timestamp = datetime.strptime(timestamp, dateTimeFormat)
        if timeDelta != None:
            timestamp += timeDelta
        timestamp = pytz.timezone(timeZone).localize(timestamp).isoformat()
        return timestamp

    def _dateFormat(timestamp: str) -> str:

        # German
        if match(r'\d{2}.\d{2}.\d{4} \d{2}:\d{2}:\d{2}.\d{1,6}', timestamp):
            return '%d.%m.%Y %H:%M:%S.%f'
        if match(r'\d{2}.\d{2}.\d{4} \d{2}:\d{2}:\d{2}', timestamp):
            return '%d.%m.%Y %H:%M:%S'
        if match(r'\d{2}.\d{2}.\d{4} \d{2}:\d{2}', timestamp):
            return '%d.%m.%Y %H:%M'
        if match(r'\d{2}.\d{2}.\d{4}', timestamp):
            return '%d.%m.%Y'

        # ISO
        if match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{1,6}\+\d{2}:\d{2}', timestamp):
            return '%Y-%m-%dT%H:%M:%S.%f+00:00'
        if match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\+\d{2}:\d{2}', timestamp):
            return '%Y-%m-%dT%H:%M:%S+00:00'
        if match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{1,6}Z', timestamp):
            return '%Y-%m-%dT%H:%M:%S.%fZ'
        if match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{1,6}', timestamp):
            return '%Y-%m-%dT%H:%M:%S.%f'
        if match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z', timestamp):
            return '%Y-%m-%dT%H:%M:%SZ'
        if match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}', timestamp):
            return '%Y-%m-%dT%H:%M:%S'
        if match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}', timestamp):
            return '%Y-%m-%dT%H:%M'

        # English I
        if match(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}.\d{1,6}\+\d{2}:\d{2}', timestamp):
            return '%Y-%m-%d %H:%M:%S.%f+00:00'
        if match(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\+\d{2}:\d{2}', timestamp):
            return '%Y-%m-%d %H:%M:%S+00:00'
        if match(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}.\d{1,6}', timestamp):
            return '%Y-%m-%d %H:%M:%S.%f'
        if match(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}', timestamp):
            return '%Y-%m-%d %H:%M:%S'
        if match(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}', timestamp):
            return '%Y-%m-%d %H:%M'
        if match(r'\d{4}-\d{2}-\d{2}', timestamp):
            return '%Y-%m-%d'

        # English II
        if match(r'\d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2}.\d{1,6}', timestamp):
            return '%Y/%m/%d %H:%M:%S.%f'
        if match(r'\d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2}', timestamp):
            return '%Y/%m/%d %H:%M:%S'
        if match(r'\d{4}/\d{2}/\d{2} \d{2}:\d{2}', timestamp):
            return '%Y/%m/%d %H:%M'
        if match(r'\d{4}/\d{2}/\d{2}', timestamp):
            return '%Y/%m/%d'

        # English III
        if match(r'\d{1,2}/\d{1,2}/\d{4} \d{2}:\d{2}:\d{2}', timestamp):
            return '%m/%d/%Y %H:%M:%S'
        if match(r'\d{1,2}/\d{1,2}/\d{4} \d{2}:\d{2}', timestamp):
            return '%m/%d/%Y %H:%M'
        if match(r'\d{1,2}/\d{1,2}/\d{4}', timestamp):
            return '%m/%d/%Y'

        return None

    def _createInstanceItemContent(content):
        itemContent = []
        for row in content:
            if row[0] not in ['unit', 'timeUnit', 'factor']:
                itemContent.append(row)

        return itemContent

    def _checkFilePath(filePath, raiseException):
        """Checks and handles the file path"""
        filePath = Path(filePath)
        if str(filePath).lower().endswith('csv'):
            if not filePath.exists():
                msg = f"File path {filePath} does not exist"
                if raiseException:
                    raise Exception(msg)
                else:
                    logger.error(msg)
                    return
            return [filePath]

        else:
            if not filePath.exists():
                msg = f"File path {filePath} does not exist"
                if raiseException:
                    raise Exception(msg)
                else:
                    logger.error(msg)
                    return
            files = [file for file in filePath.iterdir() if str(
                file).lower().endswith('csv')]
            msg = f"No csv files found in {filePath}"
            if len(files) == 0:
                if raiseException:
                    raise Exception(msg)
                else:
                    logger.error(msg)
                    return
            logger.debug(f"FileUtils._checkFilePath: {files}")
            return files

    def _readCsvFile(file, delimiter) -> list:
        with open(file) as f:
            csv_file = csv.reader(f, delimiter=delimiter)
            content = [row for row in csv_file]
            logger.debug(f"CSV file '{file.name}' read.")
        return content

    def _checkReferences(fields: list) -> list:
        """
        Checks, if reference fields have multilevel and returns a list of one-level
        fields and error an list
        """
        references = []
        errors = []
        for field in fields:
            if field.count('.') == 0:
                continue
            elif field.count('.') == 1:
                references.append(field)
            elif field.count('.') > 1:
                errors.append(field)
            else:
                pass

        return references, errors

    def _createReferenceMapping(client: object, inventoryName: str, referenceFields: list, isArray: list, content: list) -> dict:
        """
        Creates a dictionary of mapping inventoryItemIds with the import items of reference.
        """
        references = {}
        content = pandas.DataFrame(content[1:], columns=content[0])
        for field in referenceFields:
            parentField = field.split('.')[0]
            childField = field.split('.')[1]
            sysId = False
            if childField == 'sys_inventoryItemId':
                sysId = True
            if parentField in isArray:
                listOfLists = []
                for item in content[field]:
                    if len(item) == 0:
                        continue
                    listOfLists.append(
                        UtilsProperty._transformList(item, 'String'))
                nameList = list(set().union(*listOfLists))
            else:
                nameList = list(set(content[field]))

            logger.debug(f"nameList: {nameList}")

            if sysId == False:
                inventory = client.structure[inventoryName]['properties'][parentField]['inventoryName']
                df = client.items(inventory, fields=[
                                  'sys_inventoryItemId', childField], where=f'{childField} in {nameList}')
                mapping = {item[1][childField]: item[1]
                           ['sys_inventoryItemId'] for item in df.iterrows()}
            else:
                mapping = {item: item for item in nameList}

            references.setdefault(field, mapping)

        return references

    def _createParentMapping(client: object, inventoryName: str, content: list, uniqueProperty: str) -> dict:
        """
        Checks a list of fields unique fields whether they exist or not. Returns a dict of 
        value:sys_inventoryItemId. Is used to import validity Items. a value is None, if a
        sys_parentinventoryItemId does not exist.
        """
        content = pandas.DataFrame(content[1:], columns=content[0])
        uniqueValues = list(content[uniqueProperty])
        logger.debug(f"Unique value list: {uniqueValues[:4]}...")

        if uniqueProperty == 'sys_inventoryItemId':
            return {item: item for item in uniqueValues}

        where = f'{uniqueProperty} in {uniqueValues}'
        df = client.items(
            inventoryName,
            fields=[uniqueProperty, 'sys_inventoryItemId'],
            where=where
        )

        logger.debug(f"Dataframe for mapping: {df.head()}")

        if not df.empty:
            mapping = {item[1][uniqueProperty]: item[1]
                       ['sys_inventoryItemId'] for item in df.iterrows()}
            remainingUniqueValues = list(
                set(uniqueValues).difference(set(df[uniqueProperty])))
            logger.debug(
                f"Remaining unqique values: {remainingUniqueValues[:4]}...")
            if len(remainingUniqueValues) > 0:
                for item in remainingUniqueValues:
                    mapping.setdefault(item, None)
        else:
            mapping = {item: None for item in uniqueValues}

        logger.debug(f"Parent mapping: {mapping}")

        return mapping

    def _readInventoryforMapping(
            client: object,
            inventoryName: str,
            property: str,
            items: list) -> pandas.DataFrame:

        filterList = '['
        for item in items:
            filterList += f'"{item}",'
        filterList += ']'

        df = client.items(inventoryName, fields=[property, 'sys_inventoryItemId'],
                          where=f'{property} in {filterList}')
        df = df.set_index(property)
        logger.debug(f"Items to reference (first 5):\n {df.iloc[:5]}")
        return df



from itertools import islice
import pandas as pd
import pytz
from loguru import logger
from .ut import Utils
import math

gq = Utils._toGraphQL


class UtilsTimeSeries():

    def _tsCollectionToString(collection: list) -> str:
        """Converts a list of dicts with time series data into graphQL string"""

        tsData = '[\n'
        for item in collection:
            tsData += '{\n'
            for key, value in item.items():
                if key in ['sys_inventoryId', 'sys_inventoryItemId']:
                    tsData += f'  {key}: "{value}"\n'
                elif key == 'data':
                    tsData += '  data: {\n'
                    for dkey, dvalue in value.items():
                        if dkey == 'resolution':
                            tsData += f'''  resolution: {{timeUnit: {dvalue['timeUnit']}, factor: {dvalue['factor']}}}\n'''
                        elif dkey == 'unit':
                            tsData += f'''  unit: "{dvalue}"\n'''
                        elif dkey == 'dataPoints':
                            tsData += '    dataPoints: [\n'
                            for dp_key in dvalue:
                                if 'flag' not in dp_key:
                                    value = dp_key['value']
                                    if (value is None) or (isinstance(value, float) and math.isnan(value)):
                                        tsData += f'''     {{timestamp: "{dp_key['timestamp']}", value: 0 flag: MISSING}}\n'''
                                    else:
                                        tsData += f'''     {{timestamp: "{dp_key['timestamp']}", value: {value}}}\n'''
                                else:
                                    tsData += f'''     {{timestamp: "{dp_key['timestamp']}", value: {dp_key['value']}, flag: {dp_key['flag']}}}\n'''
                            tsData += '    ]\n'
                        else:
                            pass
                    tsData += '  }\n'
                else:
                    pass
            tsData += '}\n'
        tsData += ']\n'
        return tsData

    def _dataPointsToString(dataPoints: dict) -> str:
        """Converts a dictionary of datapoints to GraphQL string"""

        _dataPoints = ''
        for timestamp, value in dataPoints.items():
            if (value is None) or (isinstance(value, float) and math.isnan(value)):
                _dataPoints += f'{{ timestamp: {gq(timestamp)} value: 0 flag: MISSING }}\n'
            else:
                _dataPoints += f'{{ timestamp: {gq(timestamp)} value: {gq(value)} }}\n'

        return _dataPoints

    def _sliceDataPoints(dataPoints: dict, start: int, stop: int):
        """Return a slice of the dataPoints dictionary"""

        return dict(islice(dataPoints, start, stop))

    def _processDataFrame(
        self,
        result: dict,
        dataFrame: pd.DataFrame,
        fields: list,
        timeZone: pytz.timezone,
        displayMode: str,
        includeMissing: bool
    ) -> pd.DataFrame:
        """
        Creates a time Series DataFrame for different time series data methods
        """
        df = dataFrame
        if df.empty:
            logger.info('The query did not produce results.')
            return df
        df.loc[(df.flag == 'MISSING'), 'value'] = math.nan

        if displayMode == 'pivot':
            #pivot needs values to be set, so replace all None values with 'N/A'!
            df[fields] = df[fields].fillna('N/A')
            
            if includeMissing == False:
                df = df.pivot_table(index='timestamp', columns=fields, values='value', dropna=True)
            elif len(fields)>1:
                # Get a list of all valid combinations and pivot df
                valid_combinations = df.groupby(fields).size().reset_index()
                list_valid_combinations = list(valid_combinations[fields].itertuples(index=False, name=None))
                df = df.pivot_table(index='timestamp', columns=fields, values='value', dropna=False)
                # Remove all combinations, that don't exist
                df = df.loc[:, list_valid_combinations]
            else:
                df = df.pivot_table(index='timestamp', columns=fields, values='value', dropna=False)

            if len(df.columns) == 0:
                logger.warning(
                    f"Could not pivot DataFrame. Try different properties for fields-argument or use displayMode='rows'")
                return
            df.index = pd.to_datetime(
                df.index, format='%Y-%m-%dT%H:%M:%S').tz_convert(pytz.timezone(timeZone))
            if self.defaults.useDateTimeOffset == False:
                df.index = df.index.tz_localize(tz=None)

            columnNumber = len(result)
            dfColumnNumber = len(df.columns)
            if len(df.columns) < len(result):
                logger.info(
                    f"{columnNumber-dfColumnNumber} columns omitted due to duplicate column headers or NaN-columns.")
            return df

        elif displayMode == 'rows':
            df.set_index('timestamp', inplace=True)
            df.index = pd.to_datetime(
                df.index, format='%Y-%m-%dT%H:%M:%S').tz_convert(pytz.timezone(timeZone))
            if self.defaults.useDateTimeOffset == False:
                df.index = df.index.tz_localize(tz=None)
            return df

        else:
            Utils._error(
                self, msg=f"Unknown displayMode '{displayMode}'. Use 'pivot' or 'rows'.")
            return

    def _queryFields(fieldList: list, arrayTypeFields: list = None, arrayPageSize: int = None, filter: dict = None, recursive=False) -> str:
        """
        Transforms a Python list of fields into graphQL String of fields
        including fields of referenced inventories        
        """

        def nestedItem(item):
            nonlocal fields
            line = ''
            for i in range(itemLength - 1):
                if arrayTypeFields != None and item[i] in arrayTypeFields:
                    if filter != None and item[i] in filter.keys():
                        line += f'{item[i]} (pageSize: {arrayPageSize}, where: {filter[item[i]]}) {{ '
                    else:
                        line += f'{item[i]} (pageSize: {arrayPageSize}) {{ '
                else:
                    line += f'{item[i]} {{ '
            line += f'{item[-1]} '
            for _ in range(itemLength - 1):
                line += '}'
            line += ' \n'
            fields += line

        fields = ''
        splitList = [item.split('.') for item in fieldList]
        logger.debug(f"intermediate step - splitted list: {splitList}")

        for item in splitList:
            if len(item) == 1:
                if item == ['resolution']:
                    continue
                fields += f'{item[0]}  \n'

            else:
                itemLength = len(item)
                if recursive == False:
                    if item[-1] == '_displayValue':
                        nestedItem(item)
                    if item[-1] == 'sys_inventoryItemId':
                        nestedItem(item)
                else:
                    nestedItem(item)

        fields += f'resolution {{timeUnit, factor}}  \n'
        fields += f'_dataPeriod {{from, to}}  \n'
        return fields



# -*- coding: utf-8 -*-

import abc
import copy
from datetime import datetime, timedelta
import functools
import logging
import time
from logging.config import ConvertingDict
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from oauthlib.oauth2 import BackendApplicationClient
import requests
from requests_oauthlib import OAuth2Session
import rfc3339

from ..utils.ut_auth import AuthData

from . import const

BasicAuth = Optional[Tuple[str, str]]


class LokiEmitter(abc.ABC):
    """Base Loki emitter class."""

    success_response_code = const.success_response_code
    level_tag = const.level_tag
    logger_tag = const.logger_tag
    label_allowed_chars = const.label_allowed_chars
    label_replace_with = const.label_replace_with
    session_class = requests.Session
    auth_data: AuthData = None


    def __init__(self, url: str, tags: Optional[dict] = None, auth: BasicAuth = None, token: str = None):
        """
        Create new Loki emitter.

        Arguments:
            url: Endpoint used to send log entries to Loki (e.g. `https://my-loki-instance/loki/api/v1/push`).
            tags: Default tags added to every log record.
            auth: Optional tuple with username and password for basic HTTP authentication.

        """
        #: Tags that will be added to all records handled by this handler.
        self.tags = tags or {}
        #: Loki JSON push endpoint (e.g `http://127.0.0.1/loki/api/v1/push`)
        self.url = url
        #: Optional tuple with username and password for basic authentication.
        self.auth = auth

        #: Optional bearer token for authorization header
        self.token = token

        self._session: Optional[requests.Session] = None

    def __call__(self, record: logging.LogRecord, line: str):
        """Send log record to Loki."""
        payload = self.build_payload(record, line)
        resp = self.session.post(self.url, json=payload)
        if resp.status_code != self.success_response_code:
            raise ValueError("Unexpected Loki API response status code: {0}".format(resp.status_code))

    @abc.abstractmethod
    def build_payload(self, record: logging.LogRecord, line) -> dict:
        """Build JSON payload with a log entry."""
        raise NotImplementedError  # pragma: no cover

    @property
    def session(self) -> OAuth2Session:
        """Create HTTP session."""
        def token_updater(new_token):
            self.auth_data.token = new_token
        if self._session is None:
            self._session = OAuth2Session(self.auth_data.client_id, 
                                          token=self.auth_data.token,
                                          auto_refresh_url=self.auth_data.token_url,
                                          auto_refresh_kwargs={'client_id': self.auth_data.client_id},
                                          token_updater=token_updater)
            self._session.auth = self.auth or None
        elif self.auth_data:
            expires_at = datetime.fromtimestamp(self.auth_data.token['expires_at'])
            if expires_at - timedelta(minutes=1) < datetime.now():
                if self.auth_data.service_account_secret is None:
                    extra = { 'client_id': self.auth_data.client_id }
                    self.auth_data.token = OAuth2Session(self.auth_data.client_id, token=self.auth_data.token).refresh_token(
                        self.auth_data.token_url, **extra)
                else:
                     # login with service account returns no refresh token so we need a new login
                    app_client = BackendApplicationClient(client_id=self.auth_data.client_id)
                    self._session = OAuth2Session(client=app_client)
                    self.auth_data.token = self._session.fetch_token(token_url=self.auth_data.token_url, 
                                                            client_id=self.auth_data.client_id, 
                                                            username=self.auth_data.service_account_name, 
                                                            password=self.auth_data.service_account_secret)
                    #self._session.headers.update({'Authorization': "Bearer " + self.auth_data.token['access_token']})

        if self.token is not None:
            self._session.headers.update({'Authorization': "Bearer " + self.token})
        return self._session

    def close(self):
        """Close HTTP session."""
        if self._session is not None:
            self._session.close()
            self._session = None

    @functools.lru_cache(const.format_label_lru_size)
    def format_label(self, label: str) -> str:
        """
        Build label to match prometheus format.

        `Label format <https://prometheus.io/docs/concepts/data_model/#metric-names-and-labels>`_
        """
        for char_from, char_to in self.label_replace_with:
            label = label.replace(char_from, char_to)
        return "".join(char for char in label if char in self.label_allowed_chars)

    def build_tags(self, record: logging.LogRecord) -> Dict[str, Any]:
        """Return tags that must be send to Loki with a log record."""
        tags = dict(self.tags) if isinstance(self.tags, ConvertingDict) else self.tags
        tags = copy.deepcopy(tags)
        tags[self.level_tag] = record.levelname.lower()
        tags[self.logger_tag] = record.name

        extra_tags = getattr(record, "tags", {})
        if not isinstance(extra_tags, dict):
            return tags

        for tag_name, tag_value in extra_tags.items():
            cleared_name = self.format_label(tag_name)
            if cleared_name:
                tags[cleared_name] = tag_value

        return tags


class LokiEmitterV0(LokiEmitter):
    """Emitter for Loki < 0.4.0."""

    def build_payload(self, record: logging.LogRecord, line) -> dict:
        """Build JSON payload with a log entry."""
        labels = self.build_labels(record)
        ts = rfc3339.format_microsecond(record.created)
        stream = {
            "labels": labels,
            "entries": [{"ts": ts, "line": line}],
        }
        return {"streams": [stream]}

    def build_labels(self, record: logging.LogRecord) -> str:
        """Return Loki labels string."""
        labels: List[str] = []
        for label_name, label_value in self.build_tags(record).items():
            cleared_name = self.format_label(str(label_name))
            cleared_value = str(label_value).replace('"', r"\"")
            labels.append('{0}="{1}"'.format(cleared_name, cleared_value))
        return "{{{0}}}".format(",".join(labels))


class LokiEmitterV1(LokiEmitter):
    """Emitter for Loki >= 0.4.0."""

    def build_payload(self, record: logging.LogRecord, line) -> dict:
        """Build JSON payload with a log entry."""
        labels = self.build_tags(record)
        ns = 1e9
        ts = str(int(time.time() * ns))
        stream = {
            "stream": labels,
            "values": [[ts, line]],
        }
        return {"streams": [stream]}
